#
# gonzui - a source code search engine.
#
# Copyright (C) 2004-2005 Satoru Takabayashi <satoru@namazu.org> 
#     All rights reserved.
#     This is free software with ABSOLUTELY NO WARRANTY.
#
# You can redistribute it and/or modify it under the terms of 
# the GNU General Public License version 2.
#

require 'find'
require 'ftools'

module Gonzui
  class ExtractorError < GonzuiError; end

  class Extractor
    self.extend(Util)
    include TemporaryDirectoryUtil

    ExtnameTable = {
      ".zip"     => [ :extract_zip,    "unzip" ],
      ".tar.gz"  => [ :extract_targz,  "tar", "gunzip" ],
      ".tgz"     => [ :extract_targz,  "tar", "gunzip" ],
      ".tar.bz2" => [ :extract_tarbz2, "tar", "bunzip2" ],
      ".src.rpm" => [ :extract_srpm,   "rpm2cpio", "cpio" ]
    }

    SupportedExtnames = []
    def self.init
      ExtnameTable.each {|extname, values|
        commands = values[1, values.length - 1]
        if commands.all? {|command| command_exist?(command) }
          SupportedExtnames.push(extname)
        end
      }
    end
    init

    def self.find_archive (directory)
      Dir.entries_without_dots(directory).map {|entry|
        File.join(directory, entry)
      }.find {|file_name|
        supported_file?(file_name)
      }
    end

    def self.supported_archive? (file_name)
      return SupportedExtnames[file_name]
    end

    def self.get_archive_extname (file_name)
      ExtnameTable.keys.each {|extname|
        pattern = Regexp.new(Regexp.quote(extname) + '$') #'
        return extname if pattern.match(file_name)
      }
      return nil
    end

    def self.supported_file? (file_name)
      SupportedExtnames.include?(get_archive_extname(file_name))
    end

    def self.suppress_archive_extname (file_name)
      extname = get_archive_extname(file_name)
      pattern = Regexp.new(Regexp.quote(extname) + '$') #'
      return file_name.gsub(pattern, "")
    end

    def initialize (config, file_name)
      @archive_directory = nil
      @file_name = file_name
      @extracted_files = []
      set_temporary_directory(config.temporary_directory)
    end

    private
    def get_archive_extname (file_name)
      self.class.get_archive_extname(file_name)
    end

    def supported_file? (file_name)
      self.class.supported_file?(file_name)
    end

    def find_archive (directory)
      self.class.find_archive(directory)
    end

    def get_archiver_command_and_options (extname)
      ExtnameTable[extname]
    end

    def suppress_archive_extname (file_name)
      self.class.suppress_archive_extname(file_name)
    end

    def run_extract_command (command_line, file_name)
      command_line = sprintf("cd %s; %s", 
                             shell_escape(self.temporary_directory),
                             command_line)
      status = system(command_line)
      raise ExtractorError.new("#{file_name}: unable to extract a file") if 
        status == false
    end

    def extract_tarbz2 (file_name)
      command_line = sprintf("bunzip2 -c %s | tar xf -",
                             shell_escape(File.expand_path(file_name)))
      run_extract_command(command_line, file_name)
    end

    def extract_targz (file_name)
      command_line = sprintf("gunzip -c %s | tar xf -",
                             shell_escape(File.expand_path(file_name)))
      run_extract_command(command_line, file_name)
    end

    def extract_zip (file_name)
      command_line = sprintf("unzip -q %s",
                             shell_escape(File.expand_path(file_name)))
      run_extract_command(command_line, file_name)
    end

    def extract_srpm (file_name)
      command_line = sprintf("rpm2cpio %s | cpio -iu --quiet", 
                             shell_escape(File.expand_path(file_name)))
      run_extract_command(command_line, file_name)

      archive_file_name = find_archive(self.temporary_directory)
      raise ExtractorError.new("#{file_name}: archive not found") if 
        archive_file_name.nil?

      srpm_file_names = Dir.all_files(self.temporary_directory)
      extract_archive(archive_file_name)
      srpm_file_names.each {|file_name| File.unlink(file_name) }
    end

    def extract_archive (file_name)
      raise ExtractorError.new("#{file_name}: no such file") unless 
        File.file?(file_name)
      raise ExtractorError.new("#{file_name}: not supported archive") unless
        supported_file?(file_name)

      extname  = get_archive_extname(file_name)
      method  = ExtnameTable[extname].first
      send(method, file_name)
    end

    # Well-mannered archive doesn't scatter files when extracted.
    def has_single_directory?
      entries = Dir.entries_without_dots(self.temporary_directory)
      entries.length == 1 and 
        File.directory?(File.join(self.temporary_directory,
                                  entries.first))
    end

    # Gather scattered files in to a single directory.
    def arrange_extracted_files
      return if has_single_directory?

      entries  = Dir.entries_without_dots(self.temporary_directory)
      package_name = suppress_archive_extname(File.basename(@file_name))
      new_directory = File.join(self.temporary_directory, package_name)
      File.mkpath(new_directory)

      entries.each {|entry|
        from = File.join(self.temporary_directory, entry)
        to   = File.join(new_directory, entry)
        File.rename(from, to)
      }
    end

    def raise_error_if_empty_archive?
      entries = Dir.entries_without_dots(self.temporary_directory)
      raise ExtractorError.new("#{@file_name}: empty package") if 
        entries.length == 0
    end

    def get_archive_directory
      raise "#{@file_name}: is not well extracted" unless has_single_directory?
      entry = Dir.entries_without_dots(self.temporary_directory).first
      return File.join(self.temporary_directory, entry)
    end

    public
    def extract
      prepare_temporary_directory

      extract_archive(@file_name)
      raise_error_if_empty_archive?
      arrange_extracted_files
      return get_archive_directory
    end

    def clean
      clean_temporary_directory
    end
  end
end
