#
# gonzui - a source code search engine.
#
# Copyright (C) 2004-2005 Satoru Takabayashi <satoru@namazu.org> 
#     All rights reserved.
#     This is free software with ABSOLUTELY NO WARRANTY.
#
# You can redistribute it and/or modify it under the terms of 
# the GNU General Public License version 2.
#

require 'ftools'
require 'langscan'

module Gonzui
  include Util

  class Indexer
    @@performance_monitor = PerformanceMonitor.new

    def self.statistics 
      return "" if @@performance_monitor.empty?
      pm = @@performance_monitor
      summary = "Performance statistics:\n"
      summary << pm.heading
      summary << pm.format([Indexer, :index],
                           [Indexer, :read_content],
                           [Indexer, :add_source],
                           [Indexer, :flush_cache])
      labels = LangScan.modules.map {|m|
        [m, :scan]
      }.push([Indexer, :add_fragment])
      summary << pm.format([Indexer, :add_source], *labels)
      return summary
    end

    def initialize (config, dbm, normalized_path, content, mtime)
      @config = config
      @dbm = dbm
      @normalized_path = normalized_path
      @content = content
      @mtime = mtime

      @package_name = get_package_name
      @seqno      = 0
      @lang_id    = "\xff\xff\xff\xff" # = -1

      @wordinfo_cache = Hash.new
      @digest_cache = []

      # to be initialized
      @encoding   = nil
      @nlines     = nil
      @package_id = nil
      @path_id    = nil
      @funcfreq   = nil

      # shortcut.
      @dbm_word_cache        = @dbm.word_cache
      @dbm_word_wordid       = @dbm.word_wordid
      @dbm_wordid_word       = @dbm.wordid_word

      initialize_profilers_if_necessary
    end

    def initialize_profilers_if_necessary
      # profiler
      if @config.verbose
        @@performance_monitor.profile(Indexer, :index)
        @@performance_monitor.profile(Indexer, :read_content)
        @@performance_monitor.profile(Indexer, :add_source)
        @@performance_monitor.profile(Indexer, :add_fragment)
        @@performance_monitor.profile(Indexer, :flush_cache)
      end
    end

    def read_content
      @content, @encoding = normalize_content(@content)
      @nlines  = calc_nlines(@content)
    end

    # allow 0x09 (TAB), 0x0a (LF), 0x0c(^L), 0x0d (CR) 0x1b (ESC)
    allowed = [0x09, 0x0a, 0x0c, 0x0d, 0x1b]
    pattern = "["
    pattern << (0...0x20).find_all {|x|
      not allowed.include?(x)
    }.map {|x| sprintf("\\x%02x", x) }.join 
    pattern << "]"
    BinaryRegexp = Regexp.new(pattern)

    def binary_content? (content)
      BinaryRegexp.match(content)
    end

    def convert_to_utf8 (content)
      encoding = "ascii"
      if binary_content?(content)
        encoding = "binary"
      else 
        if @config.utf8
          content, encoding = UTF8.to_utf8(content)
        end
      end
      return content, encoding
    end

    def normalize_content (content)
      content, encoding = convert_to_utf8(content)
      unless encoding == "binary"
        content = content.untabify
        content.gsub!(/\r\n?/, "\n")
      end
      return content, encoding
    end

    def get_package_name
      parts = @normalized_path.split(File::SEPARATOR)
      if parts.length < 2
        raise GonzuiError.new("normalized path should not be flat")
      end
      return parts.first
    end

    def add_text (fragment)
      text = fragment.text
      byteno = fragment.byteno
      lineno = fragment.lineno
      type_id = Type.to_id(fragment.type)
      tokenizer = TextTokenizer.new(text)
      tokenizer.each {|word, pos, lineno_offset|
        add_word(word, byteno + pos, lineno + lineno_offset, type_id)
      }
    end

    def update_func_freq_if_necessary (fragment)
      if Type.function?(fragment.type)
        @funcfreq.increase(fragment.text)
      end
    end

    def add_fragment (fragment)
      type_id = Type.to_id(fragment.type)
      if Type.splittable?(fragment.type)
        add_text(fragment)
      else
        add_word(fragment.text, fragment.byteno, 
                 fragment.lineno, type_id)
        update_func_freq_if_necessary(fragment)
      end

      @digest_cache.push(fragment.byteno, fragment.text.length, type_id)
    end

    def flush_cache
      all_word_ids = @wordinfo_cache.keys
      all_word_ids.sort!
      all_word_ids.each {|word_id|
        @dbm.wordid_pkgids[word_id] = @package_id
        @dbm.wordpkgid_pathids[word_id + @package_id] = @path_id
        dump = DeltaDumper.dump(WordInfo, @wordinfo_cache[word_id])
        @dbm.wordpathid_info[word_id + @path_id] = dump
      }
      @dbm.pathid_wordids_dump(@path_id, all_word_ids)
      dump = DeltaDumper.dump(DigestInfo, @digest_cache)
      @dbm.pathid_digest[@path_id] = dump
    end

    def add_language (lang_abbrev, lang_name)
      if @dbm.has_lang_abbrev?(lang_abbrev)
        @lang_id = @dbm.get_lang_id(lang_abbrev)
      else
        @lang_id = make_new_lang_id
        @dbm.labbr_langid[lang_abbrev] = @lang_id
        @dbm.langid_labbr[@lang_id] = lang_abbrev
        @dbm.langid_lang[@lang_id]  = lang_name
      end
      @dbm.increase_counter(@dbm.make_nsources_key(lang_abbrev))
      @dbm.pkgid_langids[@package_id] = @lang_id
    end

    def make_new_lang_id
      @dbm.increase_counter(:nlanguages)
      return @dbm.make_new_id(:lang_id)
    end

    def make_new_package_id
      @dbm.increase_counter(:npackages)
      return @dbm.make_new_id(:package_id)
    end

    def make_new_path_id
      @dbm.increase_counter(:nfiles)
      return @dbm.make_new_id(:path_id)
    end

    def make_new_word_id
      @dbm.increase_counter(:nwords)
      return @dbm.make_new_id(:word_id)
    end

    def add_path
      assert_equal(false, @dbm.path_pathid.include?(@normalized_path))
      @path_id = make_new_path_id
      @dbm.path_pathid[@normalized_path] = @path_id
      @dbm.pathid_path[@path_id]   = @normalized_path
      @dbm.pkgid_pathids[@package_id] = @path_id
    end

    def get_fragments (scanner)
      @@performance_monitor.profile(scanner, :scan) if @config.verbose
      fragments = []
      scanner.scan(@content) {|fragment|
        fragments.push(fragment) if Type.include?(fragment.type)
      }
      fragments = fragments.sort_by {|fragment| fragment.byteno }
      return fragments
    end

    def add_source (scanner)
      @funcfreq = @dbm.make_funcfreq(scanner.lang_abbrev)
      fragments = get_fragments(scanner)
      fragments.each {|fragment|
        add_fragment(fragment)
      }
      @dbm.increase_counter(:nsources)
      @dbm.increase_counter(:nlines, @nlines)
      add_language(scanner.lang_abbrev, scanner.lang_name)
    end

    def vputs (message)
      puts message if @config.verbose
    end

    def add_source_if_necessary
      if @encoding == "binary"
        vputs "skipping BINARY: #{@normalized_path}..."
        return
      end
      scanner = LangScan.choose(@normalized_path)
      scanner = LangScan::Text if scanner.nil?
      vputs "importing #{scanner.lang_name}: #{@normalized_path}..."
      add_source(scanner)
      flush_cache
    end

    def add_word (word, byteno, lineno, type_id)
      word_id = @dbm_word_cache[word]
      if word_id.nil?
        word_id = @dbm_word_wordid[word]
        if word_id.nil?
          word_id = make_new_word_id
          @dbm_wordid_word[word_id] = word
          @dbm_word_wordid[word] = word_id
        end
        @dbm_word_cache[word] = word_id
      end

      @wordinfo_cache[word_id] ||= []
      @wordinfo_cache[word_id].push(@seqno, byteno, lineno, type_id)
      @seqno += 1
    end

    def calc_nlines (content)
      nlines = 0
      content.each_line {|line| nlines += 1 }
      return nlines
    end

    def add_package_if_necessary
      if @dbm.has_package?(@package_name)
        @package_id = @dbm.get_package_id(@package_name)
      else
        @package_id = make_new_package_id
        @dbm.pkg_pkgid[@package_name] = @package_id
        @dbm.pkgid_pkg[@package_id] = @package_name
      end
    end

    def add_file
      file_info = FileInfo.dump(@content.length, 
                                @mtime.to_i,
                                Time.now.to_i,
                                @lang_id,
                                @nlines)
      @dbm.pathid_pkgid[@path_id] = @package_id
      @dbm.set_file_content(@path_id, @content)
      @dbm.pathid_info[@path_id] = file_info
    end

    public
    def index
      read_content
      add_package_if_necessary
      add_path
      add_source_if_necessary
      add_file
    end
  end
end
