#
# gonzui - a source code search engine.
#
# Copyright (C) 2004-2005 Satoru Takabayashi <satoru@namazu.org> 
#     All rights reserved.
#     This is free software with ABSOLUTELY NO WARRANTY.
#
# You can redistribute it and/or modify it under the terms of 
# the GNU General Public License version 2.
#

require 'gonzui/delta'

module Gonzui
  module Type
    extend GetText

    class TypeGroup
      def initialize (name)
        @name = name
        @members = []
        @by_type = Hash.new
      end
      attr_reader :name

      def << (info)
        @members.push(info)
        @by_type[info.type] = info
      end

      def has_type? (type)
        @by_type.include?(type)
      end

      def each
        @members.each {|info| yield(info) }
      end

      def [] (i)
        if i.is_a?(Integer)
          @members[i]
        else
          @by_type[i]
        end
      end
    end

    class TypeID
      @@type_id = 0
      @@by_type_id = Array.new
      @@by_type = Hash.new

      def self.get_type_id (type)
        @@by_type[type]
      end

      def self.get_type (type_id)
        @@by_type_id[type_id]
      end

      def self.add (type)
        type_id = @@type_id
        @@by_type_id[type_id] = type
        @@by_type[type] = type_id
        @@type_id += 1
        return type_id
      end

      def self.has_type? (type)
        @@by_type.include?(type)
      end
    end

    class TypeInfo
      def initialize (type, name)
        @type = type
        @name = name
        @type_id = TypeID.add(type)
      end
      attr_reader :type
      attr_reader :name
      attr_reader :type_id
    end

    FunctionGroup = TypeGroup.new(N_("Function"))
    FunctionGroup << TypeInfo.new(:fundef,  N_("Function Definition"))
    FunctionGroup << TypeInfo.new(:funcall, N_("Function Call"))
    FunctionGroup << TypeInfo.new(:fundecl, N_("Function Declaration"))

    ObjectGroup = TypeGroup.new(N_("Object"))
    ObjectGroup << TypeInfo.new(:classdef,  N_("Class Definition"))
    ObjectGroup << TypeInfo.new(:moduledef, N_("Module Definition"))

    LiteralGroup = TypeGroup.new(N_("Literal"))
    LiteralGroup << TypeInfo.new(:const,     N_("Constant"))
    LiteralGroup << TypeInfo.new(:word,      N_("Word"))
    LiteralGroup << TypeInfo.new(:ident,     N_("Identifier"))
    LiteralGroup << TypeInfo.new(:integer,   N_("Integer"))
    LiteralGroup << TypeInfo.new(:floating,  N_("Floating"))
    LiteralGroup << TypeInfo.new(:imaginary, N_("Imaginary"))
    LiteralGroup << TypeInfo.new(:string,    N_("String"))
    LiteralGroup << TypeInfo.new(:character, N_("Character"))
    LiteralGroup << TypeInfo.new(:comment,   N_("Comment"))

    KeywordGroup = TypeGroup.new(N_("Keyword"))
    KeywordGroup << TypeInfo.new(:keyword, N_("Keyword"))
    KeywordGroup << TypeInfo.new(:type,    N_("Primitive Type"))

    TextGroup = TypeGroup.new(N_("Text"))
    TextGroup << TypeInfo.new(:text, N_("Text"))

    AllGroups = [
      FunctionGroup, ObjectGroup, LiteralGroup, 
      KeywordGroup, TextGroup
    ]

    module_function
    def function? (type)
      FunctionGroup.has_type?(type)
    end

    def include? (type)
      TypeID.has_type?(type)
    end

    SplittableTypes = Hash.new
    [:string, :comment, :text].each {|type| SplittableTypes[type] = true }
    def splittable? (type)
      SplittableTypes.include?(type)
    end

    def highlightable? (type)
      ! splittable?(type)
    end

    def each_group
      AllGroups.each {|group|
        yield(group)
      }
    end

    def to_id (type)
      TypeID.get_type_id(type)
    end

    def get_type (type_id)
      TypeID.get_type(type_id)
    end
  end

  module BytenoMixin
    def end_byteno
      byteno + length
    end

    def range
      byteno ... (byteno + length)
    end
  end

  module WordTypeMixin
    def type
      Type.get_type(self.type_id)
    end
  end

  class DeltaDumper
    extend Delta

    def self.dump (klass, seq)
      do_delta_list(seq, klass::DeltaSize, klass::UnitSize)
      seq.pack("w*")
    end

    def self.undump (klass, dump, &block)
      i = 0
      size = klass::UnitSize
      prev = Array.new(klass::DeltaSize, 0)
      values = []
      seq = dump.unpack("w*")
      while i < seq.length
        data = do_undelta(prev, seq[i, size]);
        if block_given?
          yield(data)
        else
          values.push(data)
        end
        i += size
      end
      return values unless block_given?
    end
  end

  WordInfo = Struct.new(:word_id, :path_id, 
                        :seqno, :byteno, :lineno, :type_id)
  class WordInfo
    include WordTypeMixin
    include BytenoMixin

    # dump info
    DeltaSize = 3
    UnitSize = 4

    def match? (target_type)
      target_type == :all or target_type == self.type
    end
  end

  DigestInfo = Struct.new(:byteno, :length, :type_id)
  class DigestInfo
    include WordTypeMixin
    include BytenoMixin

    # dump info
    DeltaSize = 1
    UnitSize = 3
  end

  FileInfo = Struct.new(:size, :mtime, :itime, :lang_id, :nlines)
  class FileInfo
    DUMP_FORMAT = "NNNa4N"

    def self.load (dump)
      self.new(*dump.unpack(DUMP_FORMAT))
    end

    def self.dump (size, mtime, itime, lang_id, nlines)
      [size, mtime, itime, lang_id, nlines].pack(DUMP_FORMAT)
    end
  end

  Occurrence = Struct.new(:byteno, :lineno, :length)
  class Occurrence
    include BytenoMixin
  end
end
