require 'langscan/c/c.so'
require 'langscan/_common'

# define Array#rindex {} for 1.8 or older 
begin
  [].rindex {}
rescue ArgumentError
  class Array
    alias clang_original_rindex rindex
    def rindex(*args)
      if block_given?
        (length-1).downto(0) {|i|
          return i if yield self[i]
        }
        nil
      else
        clang_original_rindex(*args)
      end
    end
  end
end

module LangScan
  module C
    def lang_name
      "C"
    end
    module_function :lang_name

    def lang_abbrev
      "c"
    end
    module_function :lang_abbrev

    def extnames
      [".c", ".h", ".cc", ".cpp"]
    end
    module_function :extnames

    # LangScan::C.scan iterates over C program.
    # It yields for each element which is interested by gonzui. 
    #
    # _input_ should be a String or a IO.
    def scan(input, &block)
      stack = PairStack.new
      each_fragment(input) {|f|
        stack.put(f) {|found|
          yield found
        }
      }
      stack.finish {|found|
        yield found
      }
    end
    module_function :scan

    # LangScan::C.each_fragment iterates over C-language fragments in _input_.
    # The fragments contains tokens and inter-token spaces and comments.
    #
    # _input_ can be specified as a String or a IO.
    # If a String is specified, the String itself is assumed as a C-program.
    # If a IO is specified, the content of the IO is assumed as a C-program.
    def each_fragment(input) # :yields: token
      begin
        tokenizer = Tokenizer.new(input)
        while token_info = tokenizer.get_token
          type, text, beg_lineno, beg_columnno, beg_byteno, end_lineno, end_columnno, end_byteno = token_info
          token = Fragment.new(type, text, beg_lineno, beg_byteno)
          if token.type == :ident
            if TypesHash[token.text]
              token.type = :type
            elsif KeywordsHash[token.text]
              token.type = :keyword 
            end
          end
          yield token
        end
      ensure
        tokenizer.close
      end
    end
    module_function :each_fragment

    # LangScan::C.each_token iterates over C-language tokens in _input_.
    #
    # _input_ can be specified as a String or a IO.
    # If a String is specified, the String itself is assumed as a C-program.
    # If a IO is specified, the content of the IO is assumed as a C-program.
    def each_token(input) # :yields: token
      begin
        tokenizer = Tokenizer.new(input)
        while token_info = tokenizer.get_token
          type, text, beg_lineno, beg_columnno, beg_byteno, end_lineno, end_columnno, end_byteno = token_info
          if type != :space && type != :comment
            yield Fragment.new(type, text, beg_lineno, beg_byteno)
          end
        end
      ensure
        tokenizer.close
      end
    end
    module_function :each_token

    Keywords = %w(
      auto break case char const continue default do
      double else enum extern float for goto if int
      long register return short signed sizeof static
      struct switch typedef union unsigned void volatile
      while
    )
    KeywordsHash = {}
    Keywords.each {|k| KeywordsHash[k] = k }

    Types = %w(char double float int long short void)
    TypesHash = {}
    Types.each {|k| TypesHash[k] = k }

    PairBegToEnd = {
      :punct => {
        '(' => [:paren, :punct, ')'],
        '{' => [:brace, :punct, '}'],
        '[' => [:bracket, :punct, ']'],
      },
      :preproc_beg => {
        '#' => [:preproc, :preproc_end, "\n"]
      }
    }

    PairEndToBeg = {}
    PairBegToEnd.each {|beg_type, h|
      h.each {|beg_text, (pair_type, end_type, end_text)|
        PairEndToBeg[end_type] ||= {}
        PairEndToBeg[end_type][end_text] = pair_type
      }
    }

    ArgumentSep = { :punct => { ',' => true } }

    BeforeOpenMax = 3
    AfterOpenMax = 2
    BeforeCloseMax = 2
    AfterCloseMax = 2
    TokenMax = [BeforeOpenMax, AfterOpenMax, BeforeCloseMax, AfterCloseMax].max

    # :stopdoc:
    class PairStack
      def initialize
        @stack = []
        @last_tokens = []
        @candidates = Hash.new(0)
        @opened_pairs = []
        @closed_pairs = []
      end

      def candidate_inc(token)
        @candidates[token] += 1
      end

      def candidate_dec(token, found)
        if @candidates.include? token
          if found || (@candidates[token] -= 1) == 0
            @candidates.delete token
            yield token
          end
        end
      end

      def put(fragment, &block)
        if fragment.type == :space || fragment.type == :comment
          yield fragment
          return
        end
        token = fragment

        candidate_inc token

        @opened_pairs.each {|pair| pair.after_open_tokens << token }
        if !@opened_pairs.empty? &&
           @opened_pairs[0].after_open_tokens.length == AfterOpenMax
          @opened_pairs.shift
        end

        @closed_pairs.each {|pair| pair.after_close_tokens << token }
        if !@closed_pairs.empty? &&
           @closed_pairs[0].after_close_tokens.length == AfterCloseMax
          pair = @closed_pairs.shift
          pair.check {|ct| candidate_dec(ct, true, &block) }
        end

        if (h = PairBegToEnd[token.type]) &&
           (e = h[token.text])
          pair_type, end_type, end_text = e
          before_open_tokens = @last_tokens.length <= BeforeOpenMax ?
                               @last_tokens.dup :
                               @last_tokens[-BeforeOpenMax..-1]
          pair = Pair.new(@stack.last, pair_type, before_open_tokens, token)
          pair.each_candidate_before_open {|ct| candidate_inc ct }
          @stack << pair
          @opened_pairs << pair
        elsif (h = ArgumentSep[token.type]) && h[token.text]
          @stack.last.sep_tokens << token if !@stack.empty?
        elsif (h = PairEndToBeg[token.type]) &&
              (pair_type = h[token.text]) &&
              (level = @stack.rindex {|pair| pair.pair_type == pair_type })
          unmatched_pairs = @stack.slice!(level..-1)
          matched_pair = unmatched_pairs.shift
          before_close_tokens = @last_tokens.length <= BeforeCloseMax ?
                                @last_tokens.dup :
                                @last_tokens[-BeforeCloseMax..-1]
          matched_pair.before_close_tokens = before_close_tokens
          matched_pair.close_token = token
          @closed_pairs << matched_pair
          unmatched_pairs.reverse_each {|pair|
            pair.each_candidate {|ct| candidate_dec(token, false, &block) }
          }
        end
        @last_tokens << token
        if TokenMax < @last_tokens.length
          candidate_dec(@last_tokens.shift, false, &block)
        end
      end

      def finish(&block)
        @closed_pairs.each {|pair|
          pair.check {|ct| candidate_dec(ct, true, &block) }
        }
        @candidates.keys.sort_by {|ct| ct.beg_byteno }.each {|ct| yield ct }
      end

      class Pair
        def initialize(outer, pair_type, before_open_tokens, open_token)
          @outer = outer
          @outmost = outer ? outer.outmost : self
          @pair_type = pair_type
          @beg = [open_token.type, open_token.text]
          @before_open_tokens = before_open_tokens
          @open_token = open_token
          @after_open_tokens = []
          @sep_tokens = []
          @before_close_tokens = nil
          @close_token = nil
          @after_close_tokens = []
          @candidate_direct_paren = nil
          @candidate_double_paren = nil
          @maybe_function = true
        end
        attr_reader :outer, :outmost, :pair_type, :beg
        attr_reader :before_open_tokens, :open_token, :after_open_tokens
        attr_reader :sep_tokens
        attr_accessor :before_close_tokens, :close_token
        attr_reader :after_close_tokens
        attr_accessor :maybe_function

        def each_candidate_before_open
          # name(...)
          if 1 <= @before_open_tokens.length &&
             @before_open_tokens[-1].type == :ident &&
             @pair_type == :paren
            @candidate_direct_paren = @before_open_tokens[-1]
            yield @candidate_direct_paren
          end
          # name _((...))
          if 3 <= @before_open_tokens.length &&
             @before_open_tokens[-3].type == :ident &&
             @before_open_tokens[-2].type == :ident &&
             @before_open_tokens[-1].type == :punct &&
             @before_open_tokens[-1].text == '(' &&
             @open_token.type == :punct &&
             @open_token.text == '('
            @candidate_double_paren = @before_open_tokens[-3]
            yield @candidate_double_paren
          end
        end

        def each_candidate(&block)
          yield @candidate_direct_paren if @candidate_direct_paren
          yield @candidate_double_paren if @candidate_double_paren
        end

        def check
          if name_token = @candidate_double_paren
            # name _((...))
            num_args = @sep_tokens.length + 1
            num_args = 0 if @open_token == @before_close_tokens[-1]
            if 1 <= @after_close_tokens.length &&
               @after_close_tokens[0].type == :punct &&
               @after_close_tokens[0].text == ')'
              @outer.maybe_function = true
              name_token.type = :fundecl
              yield name_token
            end
          end
          if name_token = @candidate_direct_paren
            # name(...)
            enclosed_by_fundecl_paren = @outer && outmost.pair_type == :paren
            return if enclosed_by_fundecl_paren # type id(type (*arg)());
            num_args = @sep_tokens.length + 1
            num_args = 0 if @open_token == @before_close_tokens[-1]
            if @outer &&
               @outer.pair_type == :preproc &&
               2 <= @outer.after_open_tokens.length &&
               @outer.after_open_tokens[0].type == :ident &&
               @outer.after_open_tokens[0].text == 'define' &&
               1 <= @before_open_tokens.length &&
               @outer.after_open_tokens[1] == @before_open_tokens[-1]
              if @before_open_tokens[-1].end_byteno == @open_token.beg_byteno
                # #define name(...)
                name_token.type = :fundef
                yield name_token
              end
            elsif !@outer ||
                  (!@outer.outer && # extern "C" { ... }
                   @outer.pair_type == :brace &&
                   2 <= @outer.before_open_tokens.length &&
                   @outer.before_open_tokens[-2].type == :keyword &&
                   @outer.before_open_tokens[-2].text == 'extern' &&
                   @outer.before_open_tokens[-1].type == :string &&
                   @outer.before_open_tokens[-1].text == '"C"')
              return if !@maybe_function # xxx name ((...))
              if 1 <= @after_close_tokens.length &&
                 @after_close_tokens[0].type == :punct &&
                 @after_close_tokens[0].text == ';'
                # name(...);
                name_token.type = :fundecl
                yield name_token
              elsif 1 <= @after_close_tokens.length && 
                    ((@after_close_tokens[0].type == :punct && @after_close_tokens[0].text == '{') ||
                     @after_close_tokens[0].type == :type ||
                     @after_close_tokens[0].type == :keyword ||
                     @after_close_tokens[0].type == :ident)
                # name(...) { ... }
                # name(...) int arg; { ... }
                # name(...) struct tag *arg; { ... }
                # name(...) typedefed_type arg; { ... }
                name_token.type = :fundef
                yield name_token
              else
                name_token.type = :funcall
                yield name_token
              end
            else
              return if name_token.text == 'defined' && # #if ... defined(...)
                        @outer &&
                        !@outer.outer && 
                        @outer.pair_type == :preproc &&
                        1 <= @outer.after_open_tokens.length &&
                        @outer.after_open_tokens[0].text == 'if'
              name_token.type = :funcall
              yield name_token
            end
          end

        end
      end
    end
    # :startdoc:
  end

end

if LangScan.respond_to?(:register)
  LangScan.register(LangScan::C)
end
